<?php

namespace Drupal\notes\EventSubscriber;

use Drupal\Core\Entity\EntityDefinitionUpdateManagerInterface;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityTypeEventSubscriberTrait;
use Drupal\Core\Entity\EntityTypeInterface;
use Drupal\Core\Field\FieldDefinitionListener;
use Drupal\Core\Field\FieldStorageDefinitionListener;
use Drupal\entity\BundleFieldDefinition;
use Drupal\notes\Entity\Note;
use Symfony\Component\EventDispatcher\EventSubscriberInterface;

class EntityTypeUpdateSubscriber implements EventSubscriberInterface {
  use EntityTypeEventSubscriberTrait;

  /**
   * @var \Drupal\Core\Field\FieldStorageDefinitionListener
   */
  protected $storageDefinitionListener;

  /**
   * @var \Drupal\Core\Field\FieldDefinitionListener
   */
  protected $definitionListener;

  /**
   * @var \Drupal\Core\Entity\EntityFieldManagerInterface
   */
  protected $entityFieldManager;

  /**
   * EntityTypeUpdateSubscriber constructor.
   *
   * @param \Drupal\Core\Field\FieldStorageDefinitionListener $storage_definition_listener
   * @param \Drupal\Core\Field\FieldDefinitionListener $definition_listener
   */
  public function __construct(
    FieldStorageDefinitionListener $storage_definition_listener,
    FieldDefinitionListener $definition_listener,
    EntityFieldManagerInterface $entity_field_manager
  ) {
    $this->storageDefinitionListener = $storage_definition_listener;
    $this->definitionListener = $definition_listener;
    $this->entityFieldManager = $entity_field_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function getSubscribedEvents() {
    return static::getEntityTypeEvents();
  }

  /**
   * {@inheritdoc}
   */
  public function onEntityTypeUpdate(EntityTypeInterface $entity_type, EntityTypeInterface $original) {
    if ($entity_type->get('has_notes') && !$original->get('has_notes')) {
      $this->addNoteAttachmentField($entity_type);
    }
    else if ($original->get('has_notes') && !$entity_type->get('has_notes')) {
      $this->removeNoteAttachmentField($original);
    }
    else if ($entity_type->get('has_notes') && $entity_type->get('notes_field') !== $original->get('notes_field')) {
      $this->addNoteAttachmentField($entity_type);
      // @todo: Migrate data?
      $this->removeNoteAttachmentField($original);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function onEntityTypeCreate(EntityTypeInterface $entity_type) {
    if ($entity_type->get('has_notes')) {
      $this->addNoteAttachmentField($entity_type);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function onEntityTypeDelete(EntityTypeInterface $entity_type) {
    if ($entity_type->get('has_notes')) {
      $this->removeNoteAttachmentField($entity_type);
    }
  }

  /**
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   */
  protected function addNoteAttachmentField(EntityTypeInterface $entity_type) {
    $field_name = $entity_type->get('notes_field') ?: 'notes_'.$entity_type->id();

    $storages = $this->entityFieldManager->getFieldStorageDefinitions('message');
    $field =  BundleFieldDefinition::create('entity_reference')
      ->setTargetEntityTypeId('message')
      ->setName($field_name)
      ->setLabel($entity_type->getLabel())
      ->setSetting('target_type', $entity_type->id())
      ->setSetting('notes_attachment_field', TRUE)
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE);

    if (!isset($storages[$field_name])) {
      $field->setProvider('notes');
      $this->storageDefinitionListener->onFieldStorageDefinitionCreate($field);
    }

    $fields = $this->entityFieldManager->getFieldDefinitions('message', 'note');
    if (!isset($fields[$field_name])) {
      $field->setTargetBundle('note');
      $this->definitionListener->onFieldDefinitionCreate($field);
    }
  }

  /**
   * @param \Drupal\Core\Entity\EntityTypeInterface $entity_type
   */
  protected function removeNoteAttachmentField(EntityTypeInterface $entity_type) {
    $field =  BundleFieldDefinition::create('entity_reference')
      ->setTargetEntityTypeId('message')
      ->setName($entity_type->get('notes_field') ?: 'notes_'.$entity_type->id())
      ->setLabel($entity_type->getLabel())
      ->setSetting('target_type', $entity_type->id())
      ->setSetting('notes_attachment_field', TRUE)
      ->setDisplayConfigurable('form', TRUE)
      ->setDisplayConfigurable('view', TRUE);

    $installed = $this->entityFieldManager->getFieldDefinitions('message', 'note');
    if (
      isset($installed[$field->getName()]) &&
      $installed[$field->getName()]->getFieldStorageDefinition()->getProvider() === 'notes'
    ) {
      $this->definitionListener->onFieldDefinitionDelete($field);

      $map = $this->entityFieldManager->getFieldMap();
      if (empty($map['message'][$field->getName()]['bundles'])) {
        $this->storageDefinitionListener->onFieldStorageDefinitionDelete($field);
      }
    }
  }
}
